<?php
(defined('_JEXEC') || PHP_SAPI == 'cli') or die('Restricted access');

// Checking PHP Version
if (version_compare(PHP_VERSION, '7.4.0', '<')) {
    die('Payment module "Payment system VTB" requires PHP version 7.4.0 or higher.');
}

require_once __DIR__ . '/vendor/autoload.php';

if (!class_exists('vmPSPlugin')) {
    require(JPATH_VM_PLUGINS . DS . 'vmpsplugin.php');
}

// Include the logger library
jimport('joomla.log.log');

use Vtbpay\Classes\VtbApi;

class plgVmPaymentVtbpay extends vmPSPlugin
{
    
  /**
    * Constructor of the plgVmPaymentVtbpay class. 
    * Initializes the class, setting up necessary configurations, logger, and form fields.
    * 
    * @param mixed $subject The object to observe -- event information.
    * @param array $config  An optional associative array of configuration settings.
    */
    public function __construct(& $subject, $config)
    {
        parent::__construct($subject, $config);
        $this->_loggable = true;
        $this->tableFields = array_keys($this->getTableSQLFields());
        $this->_tablepkey = $this->_tableId = 'id';
        $this->setConfigParameterable($this->_configTableFieldName, [
            'vtbpay_client_id' => ['', 'char'],
            'vtbpay_client_secret' => ['', 'char'],
            'vtbpay_test_mode' => ['', 'int'],
            'vtbpay_merchant_authorization' => ['', 'char'],
            'transaction_end_status' => ['', 'char'],
            'vtbpay_logging' => [0, 'int'],
        ]);
        JLog::addLogger(
            ['text_file' => 'plg_vmpayment_vtbpay.php'],
            JLog::ALL,
            ['plg_vmpayment_vtbpay']
        );
        JForm::addFieldPath(JPATH_PLUGINS . '/vmpayment/vtbpay/fields');
    }
    
    
    /**
     * Creates the table for this plugin if it doesn't exist.
     * Calls the method 'createTableSQL' with the name of the table to create.
     *
     * @return string SQL query to create the table.
     */
    function getVmPluginCreateTableSQL() 
    {
        return $this->createTableSQL('Payment VTB Pay Table');
    }
  
  
    /**
     * Defines the fields for the payment table.
     *
     * @return array The SQL fields for the payment table.
     */
    function getTableSQLFields() 
    {
        return [
            'id' => 'int(11) UNSIGNED NOT NULL AUTO_INCREMENT',
            'virtuemart_order_id' => 'int(1) UNSIGNED',
            'order_number' => 'char(64)',
            'virtuemart_paymentmethod_id' => 'mediumint(1) UNSIGNED',
            'payment_name' => 'char(255) NOT NULL DEFAULT \'\' ',
            'payment_order_total' => 'decimal(15,5) NOT NULL DEFAULT \'0.00000\' ',
            'payment_currency' => 'char(3) ',
            'cost_per_transaction' => 'decimal(10,2)',
            'cost_percent_total' => 'decimal(10,2)',
            'tax_id' => 'smallint(1)',
        ];
    }
    
    
    /**
     * Creates a new VtbApi instance with the given payment method configurations.
     *
     * @param object $method The payment method object.
     * @return VtbApi The new VtbApi instance.
     */
    protected function getVtbApi($method)
    {
        return new VtbApi(
            $method->vtbpay_client_id, 
            $method->vtbpay_client_secret, 
            $method->vtbpay_test_mode, 
            $method->vtbpay_merchant_authorization
        );
    }

    
    /**
     * Confirms the order in the VTB Pay payment system.
     * Creates an instance of the VTB API, sends a request to get the order link, and then handles the response.
     *
     * @param array $cart The cart with order details.
     * @param array $order The order details.
     *
     * @return mixed Returns NULL on failure, otherwise it doesn't return anything (the user gets redirected to the payment URL).
     */
    public function plgVmConfirmedOrder($cart, $order)
    {
        $order_id = $order['details']['BT']->virtuemart_order_id;
        $paymentmethod_id = $order['details']['BT']->virtuemart_paymentmethod_id;
        $email = $order['details']['BT']->email;
        $order_total = $order['details']['BT']->order_total;
        $return_url = JURI::root() . 'index.php?option=com_virtuemart&view=pluginresponse&task=pluginresponsereceived&on=' . $order['details']['BT']->order_number . '&pm=' . $paymentmethod_id;

        try {    
            if (!($method = $this->getVmPluginMethod($paymentmethod_id))) {
                throw new \Exception(JText::_('PLG_VMPAYMENT_VTBPAY_NO_METHOD_FOUND'));
            }
            
            if (!$this->selectedThisElement($method->payment_element)) {
                return NULL;
            }

            // Once the payment has been processed (but possibly before the final status is known),
            // store the initial transaction details in the database
            $this->storePaymentData($method, $order);
          
            $response = $this->getVtbApi($method)->getOrderLink(
                $order_id, 
                $email, 
                time(), 
                $order_total, 
                $return_url
            );
            
            if($method->vtbpay_logging) $this->vtbpayLogger(__FUNCTION__ . ' > getOrderLink. ', JLog::DEBUG, [
                'arguments' => [
                    'order_id' => $order_id, 
                    'email' => $email,
                    'time' => time(), 
                    'order_total' => $order_total, 
                    'return_url' => $return_url
                ],
                'response' => $response
            ]);

            if (
                isset($response['object']['payUrl']) &&
                !empty($pay_url = $response['object']['payUrl'])
            ) {
                // Handle successful request
                JFactory::getApplication()->redirect($pay_url);
            }
            else {
                throw new \Exception(JText::_('PLG_VMPAYMENT_VTBPAY_URL_FETCH_ERROR'));
            }
            
        } catch (\Exception $e) {
            $msg = $e->getMessage() . ' Order ID: ' . $order_id;
            // Handle exception and log error
            $this->vtbpayLogger(sprintf(__FUNCTION__ . ' > VtbPay Exception: %s', $msg), JLog::ERROR);
            JFactory::getApplication()->enqueueMessage(JText::_('PLG_VMPAYMENT_VTBPAY_ERROR') . ': ' . $msg, 'error');
        }

        return FALSE;
    }
    
    
    /**
     * Store payment data in the database
     *
     * @param array $data Payment data
     * @param object $order Order details
     * @return void
     */
    protected function storePaymentData($data, $order)
    {
        // Store the data using the storePSPluginInternalData method from the parent class
        $this->storePSPluginInternalData([
            'virtuemart_order_id' => $order['details']['BT']->virtuemart_order_id,
            'order_number' => $order['details']['BT']->order_number,
            'virtuemart_paymentmethod_id' => $data->virtuemart_paymentmethod_id,
            'payment_name' => $this->renderPluginName($data),
            'payment_order_total' => $order['details']['BT']->order_total,
            'payment_currency' => $this->getPaymentCurrency($data),
            'cost_per_transaction' => $data->cost_per_transaction,
            'cost_percent_total' => $data->cost_percent_total,
            'tax_id' => $data->tax_id
        ]);
    }
    
    
    /**
     * Handles the response from VTB Pay after payment is made.
     * Verifies the order, the payment method, and the order number, then checks the payment status.
     *
     * @param string $html The HTML to output.
     *
     * @return bool Returns true if payment is successful and response is handled, otherwise logs an error and returns false.
     */
    public function plgVmOnPaymentResponseReceived(&$html)
    {
        try {
            $input = JFactory::getApplication()->input;
            $virtuemart_paymentmethod_id = $input->getInt('pm');
            
            if (!$this->selectedThisByMethodId($virtuemart_paymentmethod_id)) {
                return NULL; // Another method was selected, do nothing
            }
            
            $method = $this->getVmPluginMethod($virtuemart_paymentmethod_id);

            if (!$method) {
                throw new \Exception(JText::_('PLG_VMPAYMENT_VTBPAY_NO_METHOD_FOUND'));
            }

            $order_number = $input->getString('on');
            if (!$order_number) {
                throw new \Exception(JText::_('PLG_VMPAYMENT_VTBPAY_NO_ORDER_NUMBER_FOUND'));
            }
            
            $order_id = $input->getString('orderId');
            if (!$order_id) {
                throw new \Exception(JText::_('PLG_VMPAYMENT_VTBPAY_NO_ORDER_ID_FOUND'));
            }

            if (!class_exists('VirtueMartModelOrders')) {
                require(JPATH_VM_ADMINISTRATOR . DS . 'models' . DS . 'orders.php');
            }

            $order_model = VmModel::getModel('orders');
            $order = $order_model->getOrder($order_id);

            if (!$order) {
                throw new \Exception(JText::_('PLG_VMPAYMENT_VTBPAY_NO_ORDER_FOUND') . ' Order ID: ' . $order_id);
            }
                                     
            $response = $this->getVtbApi($method)->getOrderInfo($order_id);
            
            if($method->vtbpay_logging) $this->vtbpayLogger(__FUNCTION__ . ' > getOrderInfo. ', JLog::DEBUG, [
                'arguments' => [
                    'order_id' => $order_id
                ],
                'response' => $response
            ]);

            if (
                !isset($response['object']['status']['value']) || 
                $response['object']['status']['value'] !== 'PAID'
            ) {
                throw new \Exception(JText::_('PLG_VMPAYMENT_VTBPAY_PAYMENT_FAILURE') . ' Order ID: ' . $order_id);
            }
            
            // Handle successful request
            $this->initiateAsPaid($order, $order_model, $method, $order_id);
            
            return true;
            
        } catch (\Exception $e) {
            // Handle exception and log error
            $this->vtbpayLogger(sprintf(__FUNCTION__ . ' > VtbPay Exception: %s', $e->getMessage()), JLog::ERROR);
            JFactory::getApplication()->enqueueMessage(JText::_('PLG_VMPAYMENT_VTBPAY_ERROR') . ': ' . $e->getMessage(), 'error');
        }

        return false;
    }
    
    
    /**
     * Initiates order as paid and updates its status.
     *
     * This function sets the order status to the specified transaction end status,
     * notifies the customer, adds a comment to the order indicating that it has been paid
     * with the "Payment system VTB," clears order-related session data, and empties the cart.
     *
     * @param array $order The order details.
     * @param object $order_model The orders model object.
     * @param object $method The payment method object.
     * @param int $order_id The ID of the order.
     * @return void
     */
    private function initiateAsPaid($order, $order_model, $method, $order_id) {
        $order['order_status'] = $method->transaction_end_status; 
        $order['virtuemart_order_id'] = $order_id;
        $order['customer_notified'] = 1;
        $order['comments'] = 'Order has been paid with "Payment system VTB".';
        $order_model->updateStatusForOneOrder($order_id, $order, true);
        JFactory::getApplication()->enqueueMessage(JText::_('PLG_VMPAYMENT_VTBPAY_PAYMENT_SUCCESS'));
        
        // Remove the order details from the session
        JFactory::getSession()->clear('order_number');
        JFactory::getSession()->clear('pm');
        
        // Clear the cart
        if (!class_exists('VirtueMartCart')) {
            require(JPATH_VM_SITE . DS . 'helpers' . DS . 'cart.php');
        }
        $cart = VirtueMartCart::getCart();
        $cart->emptyCart();
    }


    /**
     * Displays the payment information in the front-end when showing an order.
     *
     * @param int $virtuemart_order_id The ID of the order
     * @param int $virtuemart_paymentmethod_id The ID of the payment method
     * @param string &$payment_name The name of the payment method
     * @return void
     */
    public function plgVmOnShowOrderFEPayment($virtuemart_order_id, $virtuemart_paymentmethod_id, &$payment_name)
    {
        $this->onShowOrderFE($virtuemart_order_id, $virtuemart_paymentmethod_id, $payment_name);
    }


    /**
     * Checks the input data of the payment method within the checkout process.
     *
     * @param object &$cart The cart object
     * @return bool Always true since no data input is required for this payment method
     */
    public function plgVmOnCheckoutCheckDataPayment(&$cart) {
        // No data input required for this payment method
        return true;
    }


    /**
     * Adds the input data of the payment method to the cart, after the payment method is selected.
     *
     * @param object &$cart The cart object
     * @param string &$msg The message to be displayed (if any)
     * @return bool Always true since no data input is required for this payment method
     */
    public function plgVmOnSelectCheckPayment(&$cart, &$msg) {
        // No data input required for this payment method
        return true;
    }
    
    
    /**
     * Displays the plugin methods in the cart, typically for editing shipment/payment details.
     *
     * @param VirtueMartCart $cart Cart object
     * @param int $selected ID of the selected method
     * @param string &$htmlIn The HTML string being constructed
     * @return bool|null True on success, false on failure, null when this plugin was not selected
     */
    public function plgVmDisplayListFEPayment(VirtueMartCart $cart, $selected = 0, &$htmlIn)
    {
        return $this->displayListFE($cart, $selected, $htmlIn);
    }
    
    
    /**
     * Calculates the price (value, tax_id) of the selected payment method. This method is called by the calculator.
     *
     * @param VirtueMartCart $cart The current cart
     * @param array &$cart_prices The new cart prices
     * @param string &$cart_prices_name The name of the cart prices to be calculated
     * @return bool|null|false True if successful, false if the shipping rate is not valid anymore, null if the method was not selected
     */
    public function plgVmonSelectedCalculatePricePayment(VirtueMartCart $cart, array &$cart_prices, &$cart_prices_name)
    {
        return $this->onSelectedCalculatePrice($cart, $cart_prices, $cart_prices_name);
    }
    
    
    /**
     * Checks how many payment plugins are available. If only one, the user will not have the choice. Enters edit_xxx page.
     *
     * @param VirtueMartCart $cart The cart object
     * @param array $cart_prices Cart prices array
     * @param int &$paymentCounter The payment counter
     * @return int|null 0 if more than one plugin was found, virtuemart_xxx_id if only one plugin is found, null if no plugin was found
     */
    function plgVmOnCheckAutomaticSelectedPayment(VirtueMartCart $cart, array $cart_prices = [], &$paymentCounter)
    {
        return $this->onCheckAutomaticSelected($cart, $cart_prices, $paymentCounter);
    }
    
    
    /**
     * This method reimplements vmPaymentPlugin::checkPaymentConditions() to check if the cart's total sales price 
     * is within the acceptable range for the active payment method.
     *
     * @param VirtueMartCart $cart Cart object containing all the items and their details in the cart
     * @param object $activeMethod The active payment method object, containing its parameters such as min_amount and max_amount
     * @param array $cart_prices An associative array containing all the price details related to the cart, including 'salesPrice'
     * @return bool Returns true if the total sales price is within the acceptable range for the active payment method, false otherwise
     */
    function checkConditions($cart, $activeMethod, $cart_prices)
    {
        $this->convert($activeMethod);
        $amount = $cart_prices['salesPrice'];
        return (($amount >= $activeMethod->min_amount && $amount <= $activeMethod->max_amount)
            || ($amount >= $activeMethod->min_amount && empty($activeMethod->max_amount)));
    }
    

    /**
     * This method is used to convert the min_amount and max_amount properties of the provided object to float.
     *
     * @param object $method The object containing min_amount and max_amount properties
     * @return void
     */
    function convert($method)
    {
        $method->min_amount = (float) $method->min_amount;
        $method->max_amount = (float) $method->max_amount;
    }
    
    
    /**
     * Display stored payment data for an order in the backend.
     *
     * @param int $virtuemart_order_id The ID of the order
     * @param int $virtuemart_payment_id The ID of the payment
     * @return string|null The HTML containing the order's payment data if available, otherwise NULL
     */
    function plgVmOnShowOrderBEPayment($virtuemart_order_id, $virtuemart_payment_id) {

        if (!$this->selectedThisByMethodId($virtuemart_payment_id)) {
            return NULL; // Another method was selected, do nothing
        }

        if (!($paymentTable = $this->getDataByOrderId($virtuemart_order_id))) {
            return NULL;
        }
        vmLanguage::loadJLang('com_virtuemart');
        
        if ($paymentTable->payment_currency == 0) $paymentTable->payment_currency = '₽';

        $html = '<table class="adminlist table">' . "\n";
        $html .= $this->getHtmlHeaderBE();
        $html .= $this->getHtmlRowBE('COM_VIRTUEMART_PAYMENT_NAME', $paymentTable->payment_name);
        $html .= $this->getHtmlRowBE('PLG_VMPAYMENT_VTBPAY_TOTAL', $paymentTable->payment_order_total . ' ' . $paymentTable->payment_currency);
        $html .= '</table>' . "\n";
        return $html;
    }


    /**
     * Handles the printing of an Order's payment details.
     *
     * @param int $order_number The order number
     * @param int $method_id The ID of the payment method used
     * @return mixed The output of the onShowOrderPrint method
     */
    public function plgVmOnShowOrderPrintPayment($order_number, $method_id) 
    {
        return $this->onShowOrderPrint($order_number, $method_id);
    }


    /**
     * Stores the payment info into the order.
     *
     * @param int $jplugin_id The ID of the Joomla plugin
     * @return mixed The output of the onStoreInstallPluginTable method
     */
    public function plgVmOnStoreInstallPaymentPluginTable($jplugin_id) 
    {
        return $this->onStoreInstallPluginTable($jplugin_id);
    }
    
    
    /**
     * Sets the parameters on the payment plugin table.
     *
     * @param string $name The name of the plugin
     * @param int $id The ID of the plugin
     * @param object &$table The table to set the parameters on
     * @return mixed The result of the setOnTablePluginParams function
     */
    function plgVmSetOnTablePluginParamsPayment($name, $id, &$table)
    {
        return $this->setOnTablePluginParams($name, $id, $table);
    }
    
    
    /**
     * Declare the payment plugin parameters for VirtueMart 3.
     *
     * @param array &$data The data array containing the parameters
     * @return mixed The result of the declarePluginParams function
     */
    function plgVmDeclarePluginParamsPaymentVM3(&$data)
    {
        return $this->declarePluginParams('payment', $data);
    }
    
    
    /**
     * Declare the payment plugin parameters.
     *
     * @param string $name The name of the plugin
     * @param int $id The ID of the plugin
     * @param array &$data The data array containing the parameters
     * @return mixed The result of the declarePluginParams function
     */
    function plgVmDeclarePluginParamsPayment($name, $id, &$data) 
    {
  		  return $this->declarePluginParams('payment', $name, $id, $data);
  	}
    
    
    /** 
     * Logs a message with optional context. Log type can be either DEBUG or ERROR.
     * 
     * @param string $message The message to log
     * @param string $type The type of log. Can be either DEBUG or ERROR
     * @param array $context An optional array of context data to log with the message
     * @param bool $print_r If set to true, the context array will be printed using print_r instead of json_encode
     * @return void
     */
    private function vtbpayLogger(
        $message, 
        $type = '',
        $context = [],
        $print_r = false
    ) {
				if (empty($type)) $type = JLog::DEBUG;
        
        if (!empty($context) && is_array($context)) {
            $message .= ' Context: '. PHP_EOL;
            if ($print_r) {
                $message .= print_r($context, true) . PHP_EOL;
            }
            else {
                $message .= json_encode($context);
            }
        }
        
        JLog::add($message, $type, 'plg_vmpayment_vtbpay');
    }
}
